from pyspark.sql.types import UserDefinedType, StringType
from decimal import *


class BigNumericUDT(UserDefinedType):
    """
    SQL user-defined type (UDT) for BigNumeric.
    """

    @classmethod
    def sqlType(cls):
        return StringType()

    @classmethod
    def module(cls):
        return "google.cloud.spark.bigquery.big_numeric_support"

    @classmethod
    def scalaUDT(cls):
        return "org.apache.spark.bigquery.BigNumericUDT"

    def serialize(self, obj):
        if isinstance(obj, BigNumeric):
            return str(obj.number)
        else:
            raise TypeError("cannot serialize %r of type %r" % (obj, type(obj)))

    def deserialize(self, datum):
        return BigNumeric(Decimal(str(datum)))

    def simpleString(self):
        return "BigNumeric"


class BigNumeric:
    def __init__(self, number):
        if isinstance(number, Decimal):
            self.number = number
        else:
            raise TypeError(
                "BigNumeric Constuctor: expecting Decimal but got %r " % type(number)
            )



import requests
import json


def send_teams_webhook(
    webhook_url,
    main_title="",
    main_msg="",
    activity_title="",
    activity_msg="",
    theme_color="#0000FF",
):
    theme_color = theme_color
    # teams_webhook_url = "https://shinsegaegroup.webhook.office.com/webhookb2/f6ed8cf6-f350-482a-aa1d-ed7fd262d466@d4ffc887-d88d-41cc-bf6a-6bb47ec0f3ca/IncomingWebhook/2868210da0d24639b1413c093bf3da9a/ca7346bc-3396-46fe-9cba-7a9c18460e7d"

    headers = {"Content-Type": "application/json"}

    message_to_send = {
        "title": main_title,
        "text": main_msg,
        "themeColor": theme_color,
        "sections": [{"activityTitle": activity_title, "text": activity_msg}],
    }
    response = requests.post(webhook_url, json=message_to_send, headers=headers)

    if not response.status_code == 200:
        print(f"Send Webhook Failed!: {response.status_code}, {response.text}")



import json
from urllib.parse import unquote_plus
import requests
from datetime import datetime


webhook_msg = []
main_msg = []


def run_nes_api(notebook_params):
    global webhook_msg

    nes_url = "https://nes.shinsegae.ai/v1/runs"

    nes_headers = {
        "Content-Type": "application/json",
        "accept": "application/json",
    }

    nes_params = {
        "input_url": "https://github.com/emartdt/datafabric-notebooks/blob/main/etl/autoload/s3_to_bq_autoload.ipynb",
        "parameters": notebook_params,
        "runtime": "",
        "host_network": True,
    }
    json_params = json.dumps(nes_params)

    response = requests.post(nes_url, headers=nes_headers, data=json_params)

    if response.status_code == 200 | 201:
        response_data = response.json()
        webhook_msg.append(f"[Commuter]({response_data.get('output_url')})")
        return True
    else:
        webhook_msg.append(
            f"NES Request failed with status code : {response.status_code}"
        )
        response_data = response.content
        webhook_msg.append(f"Response MSG : {response_data}")
        return False


def send_teams_webhook(webhook_url):
    global main_msg, webhook_msg

    default = "https://shinsegaegroup.webhook.office.com/webhookb2/f6ed8cf6-f350-482a-aa1d-ed7fd262d466@d4ffc887-d88d-41cc-bf6a-6bb47ec0f3ca/IncomingWebhook/2868210da0d24639b1413c093bf3da9a/ca7346bc-3396-46fe-9cba-7a9c18460e7d"

    teams_webhook_url = webhook_url if webhook_url else default

    current = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    main_msg.append(f"* AWS Lambda Runtime : {current}")

    headers = {"Content-Type": "application/json"}

    message_to_send = {
        "title": "[AWS to BigQuery Autoload Report]",
        "text": "\n".join(main_msg),
        "themeColor": "#0000FF",
        "sections": [{"activityTitle": "NES 실행 결과", "text": "\n".join(webhook_msg)}],
    }
    response = requests.post(teams_webhook_url, json=message_to_send, headers=headers)

    if not response.status_code == 200:
        print(f"Send Webhook Failed!: {response.status_code}, {response.text}")


def isAllowedExtension(extension):
    global webhook_msg

    # parquet의 경우 해당 폴더에 _SUCCESS 파일 생성 시 '*.parquet'으로 Read
    allowed_extensions = ["csv", "json", "_success"]
    if extension.lower() in allowed_extensions:
        return True
    else:
        if extension.lower() == "schema":
            webhook_msg.append(f"Schema File Uploaded.")
        else:
            webhook_msg.append(f"Validate file format. Input format : {extension}\n")
        return False
    # raise Exception("This extension is not supported : {}, Only support csv/json/parquet".format(extension))


def s3_autoload(event, webhook_url=""):
    global webhook_msg, main_msg

    bucket = event["Records"][0]["s3"]["bucket"]["name"]
    key = unquote_plus(event["Records"][0]["s3"]["object"]["key"], encoding="utf-8")

    print(f">>> Call Lambda Handler : {key}")
    paths = key.split("/")
    project = "emart-datafabric"
    dataset = "temp_7d"
    table = ".".join(paths[-1].split(".")[0:-1])

    file = paths[-1]
    extension = file.split(".")[-1]

    if len(paths) > 2:
        project = paths[0]
        dataset = paths[1]
        table = paths[2]

    user_ip = event["Records"][0]["requestParameters"]["sourceIPAddress"]
    size = event["Records"][0]["s3"]["object"]["size"]
    timeCreated = event["Records"][0]["eventTime"]
    eventName = event["Records"][0]["eventName"]

    # TODDO: Make: eventName 기준으로 update면 updated 사용
    data = {
        "bucket": bucket,
        "key": key,
        "project": project,
        "dataset": dataset,
        "table": table,
        "file": file,
        "extension": extension,
        "user_ip": user_ip,
        "size": f"{size}",
        "timeCreated": timeCreated,
        "updated": timeCreated,
        "eventName": eventName,
    }

    for key, value in data.items():
        main_msg.append(f"* {key} : {value}\n")

    try:
        print(json.dumps(data, indent=2))

        if isAllowedExtension(extension):
            run_nes_api(data)
            send_teams_webhook(webhook_url)

        # Cloudfunction에서 메시지 중첩되어 보냄. 그래서 초기화 코드 추가
        webhook_msg = []
        main_msg = []
        return True
    except Exception as e:
        print(e)
        raise e



from google.cloud import bigquery
from google.cloud.exceptions import NotFound
from pydatafabric.vault_utils import get_secrets
from pydatafabric.gcp import get_bigquery_client
import pandas as pd
import cx_Oracle

GCP_PROJECT = "smart-ruler-304409"
TEMP_DATASET = "temp_1d"
ORACLE_SECRET = None
DCRYPTO_LIST = pd.DataFrame([])


def set_oracle_secret(secret):
    global ORACLE_SECRET
    ORACLE_SECRET = secret


def set_dcrypto_list(dcrypto_list):
    global DCRYPTO_LIST
    DCRYPTO_LIST = dcrypto_list


def get_enc_type(owner, table_name, column_name):
    global DCRYPTO_LIST
    enc_type = None
    enc_list = None
    try:
        enc_list = DCRYPTO_LIST.query(
            "(OWNER==@owner) and (TABLE_NAME==@table_name) and (COLUMN_NAME==@column_name)"
        )[["ENC_TYPE"]]
        if len(enc_list.index) == 1:
            enc_type = enc_list.iloc[0]["ENC_TYPE"]
    except NameError:
        enc_type = None

    return enc_type


def get_enc_list(owner, table_name):
    global DCRYPTO_LIST
    try:
        enc_list = DCRYPTO_LIST.query("(OWNER==@owner) and (TABLE_NAME==@table_name)")[
            ["COLUMN_NAME", "ENC_TYPE"]
        ]
    except NameError:
        enc_list = None

    return enc_list


def get_oracle_connection():
    connection = cx_Oracle.connect(
        user=ORACLE_SECRET["user"],
        password=ORACLE_SECRET["password"],
        dsn=ORACLE_SECRET["cx_url"],
        encoding=ORACLE_SECRET["encoding"],
        nencoding=ORACLE_SECRET["nencoding"],
    )

    return connection


def get_oracle_schema(owner, table_name):
    schema_sql = """
    SELECT A.OWNER AS SCHEMA_NAME, A.TABLE_NAME AS TABLE_NAME, E.COMMENTS AS TABLE_COMMENTS,
        CASE WHEN C.COLUMN_NAME IS NOT NULL THEN '1' ELSE NULL END AS IS_PK,
        B.COLUMN_NAME AS COLUMN_NAME, D.COMMENTS AS COLUMN_COMMENTS, B.COLUMN_ID AS POSITION,
        B.DATA_TYPE AS DATA_TYPE_NAME, B.NULLABLE AS IS_NULLABLE,
        B.DATA_LENGTH AS LENGTH,
        B.DATA_SCALE AS SCALE, B.DATA_PRECISION AS PRECISION
    FROM ALL_TABLES A
        INNER JOIN ALL_TAB_COLUMNS B
        ON  A.TABLE_NAME = B.TABLE_NAME
        INNER JOIN ALL_COL_COMMENTS D
        ON B.TABLE_NAME = D.TABLE_NAME
        AND B.COLUMN_NAME = D.COLUMN_NAME
        LEFT JOIN (
            SELECT A.TABLE_NAME, B.COLUMN_NAME
                FROM ALL_CONSTRAINTS A
                    INNER JOIN ALL_CONS_COLUMNS B
                    ON A.CONSTRAINT_NAME = B.CONSTRAINT_NAME
                    AND A.CONSTRAINT_TYPE = 'P'
        ) C
        ON B.TABLE_NAME = C.TABLE_NAME
        AND B.COLUMN_NAME = C.COLUMN_NAME
        INNER JOIN ALL_TAB_COMMENTS E
        ON  A.TABLE_NAME = E.TABLE_NAME
    WHERE A.OWNER IN (:owner)
    AND A.TABLE_NAME = :table_name
    ORDER BY A.TABLE_NAME, B.COLUMN_ID
    """

    bind_param = dict(owner=owner.upper(), table_name=table_name.upper())
    oracle_schema = oracle_to_pandas(
        owner=owner, table_name=table_name, sql=schema_sql, bind_param=bind_param
    )

    return oracle_schema


def output_type_handler(cursor, name, default_type, size, precision, scale):
    import datetime as dt
    import decimal as dc

    if default_type in (
        cx_Oracle.DB_TYPE_VARCHAR,
        cx_Oracle.DB_TYPE_CHAR,
        cx_Oracle.DB_TYPE_NCHAR,
        cx_Oracle.DB_TYPE_NVARCHAR,
    ):
        return cursor.var(str, arraysize=cursor.arraysize, bypass_decode=True)
    elif default_type == cx_Oracle.DB_TYPE_CLOB:
        return cursor.var(
            cx_Oracle.DB_TYPE_LONG, arraysize=cursor.arraysize, bypass_decode=True
        )
    elif default_type in (
        cx_Oracle.DB_TYPE_DATE,
        cx_Oracle.DB_TYPE_TIMESTAMP,
        cx_Oracle.DB_TYPE_TIMESTAMP_TZ,
        cx_Oracle.DB_TYPE_TIMESTAMP_LTZ,
    ):
        return cursor.var(str, arraysize=cursor.arraysize, bypass_decode=True)
    elif default_type == cx_Oracle.DB_TYPE_NUMBER:
        return cursor.var(dc.Decimal, arraysize=cursor.arraysize, bypass_decode=False)


def make_dict_factory(cursor):
    columnNames = [row[0] for row in cursor.description]

    def createRow(*args):
        return dict(zip(columnNames, args))

    return createRow


def getDefault(val, default):
    return val if val is not None else default


def convert_oracle_to_bigquery_schema(owner, table_name, partition_field=None):
    oracle_schema = get_oracle_schema(owner, table_name)
    schema = []
    if partition_field and isinstance(partition_field, bigquery.SchemaField):
        schema.append(partition_field)

    for i in range(len(oracle_schema)):
        row = oracle_schema.iloc[i]
        description = getDefault(row["column_comments"], "")
        column = row["column_name"].lower()
        oc_type = row["data_type_name"]
        nullable = row["is_nullable"]
        scale = getDefault(row["scale"], "")
        precision = getDefault(row["precision"], "")
        length = row["length"]

        bq_type = "STRING"
        if oc_type in ("VARCHAR2", "CHAR", "NVARCHAR2", "CLOB"):
            bq_type = "STRING"
        elif oc_type == "NUMBER":
            if precision and precision >= 0:
                if scale == 0:
                    if precision < 9:
                        bq_type = "INTEGER"
                    elif precision >= 9 and precision < 22:
                        bq_type = "NUMERIC"
                    elif precision >= 22:
                        bq_type = "BIGNUMERIC"
                elif scale > 0:
                    bq_type = "FLOAT"
            else:
                bq_type = "NUMERIC"
        elif oc_type == "LONG":
            bq_type = "NUMERIC"
        elif oc_type == "DATE":
            bq_type = "STRING"
        elif oc_type.startswith("TIMESTAMP"):
            bq_type = "DATETIME"

        mode = "REQUIRED" if nullable == "N" else "NULLABLE"

        schema.append(
            bigquery.SchemaField(column, bq_type, mode=mode, description=description)
        )

    return schema


def decode_byte(objs, encode_type="cp949"):
    if type(objs) is list:
        result = objs
        for i in range(len(result)):
            for col in result[i].keys():
                if isinstance(result[i][col], bytes):
                    result[i][col] = result[i][col].decode(encoding=encode_type)
        return result
    elif type(objs) is tuple:
        temp = list(objs)
        result = []
        for item in temp:
            if isinstance(item, bytes):
                try:
                    result.append(item.decode(encoding=encode_type))
                except UnicodeDecodeError:
                    lens = len(item) - 1
                    result.append(item[:lens].decode(encoding=encode_type))
            else:
                result.append(item)
        return result
    else:
        return objs


def decrypt_columns(dataframe, enc_list, columns):
    import hashlib

    result = dataframe

    for row in enc_list.itertuples():
        enc_column_name = row.COLUMN_NAME

        if enc_column_name in columns:
            result[enc_column_name] = result[enc_column_name].apply(
                lambda x: hashlib.sha256(x.encode()).hexdigest()
                if (pd.notnull(x))
                else x
            )

    return result


def oracle_fetchall_to_pandas(
    owner, table_name, sql=None, bind_param=None, enc_list=None
):
    result = None

    if sql is None:
        sql = "SELECT * FROM {}".format(table_name)
    else:
        sql = sql

    connection = get_oracle_connection()
    cursor = connection.cursor()
    cursor.outputtypehandler = output_type_handler
    cursor.execute(
        "ALTER SESSION SET NLS_DATE_FORMAT = 'YYYY-MM-DD HH24:MI:SS' NLS_TIMESTAMP_TZ_FORMAT = 'YYYY-MM-DD HH24:MI:SS.FF3'"
    )

    if bind_param is None:
        cursor.execute(sql)
    else:
        cursor.execute(sql, bind_param)

    columnNames = [row[0] for row in cursor.description]
    cursor.rowfactory = make_dict_factory(cursor)
    result = cursor.fetchall()
    result = decode_byte(result)

    result = pd.DataFrame(data=result)

    if enc_list:
        result = decrypt_columns(result, enc_list, columnNames)

    cursor.close()
    connection.close()

    return result


def oracle_to_pandas(owner, table_name, sql=None, bind_param=None, enc_list=None):
    result = None

    if sql is None:
        sql = "SELECT * FROM {}".format(table_name)
    else:
        sql = sql

    connection = get_oracle_connection()
    cursor = connection.cursor()
    cursor.outputtypehandler = output_type_handler
    cursor.execute(
        "ALTER SESSION SET NLS_DATE_FORMAT = 'YYYY-MM-DD HH24:MI:SS' NLS_TIMESTAMP_TZ_FORMAT = 'YYYY-MM-DD HH24:MI:SS.FF3'"
    )

    if bind_param is None:
        cursor.execute(sql)
    else:
        cursor.execute(sql, bind_param)

    columnNames = [row[0] for row in cursor.description]
    result = map(lambda x: decode_byte(x), cursor)

    result = pd.DataFrame(data=result, columns=columnNames)

    if enc_list:
        result = decrypt_columns(result, enc_list, columnNames)

    result = result.rename(columns=str.lower)
    cursor.close()
    connection.close()

    return result


def bq_dataset_exists(dataset, project=GCP_PROJECT):
    try:
        bq = get_bigquery_client(project=project)
        dataset = bq.get_dataset(dataset)
        bq.close()
    except NotFound:
        return False
    return True


def create_bq_table_with_oracle_schema(
    owner,
    oracle_table,
    bq_dataset,
    bq_project=GCP_PROJECT,
    bq_table=None,
    partition_type=None,
    partition_field=None,
):
    if not bq_dataset_exists(bq_dataset, bq_project):
        print("Not Exists Dataset: {}".format(bq_dataset))
        return False

    bq_table_name = bq_table if bq_table else oracle_table.lower()

    if partition_field:
        partition_field_schema = bigquery.SchemaField(
            partition_field, "DATE", mode="REQUIRED", description="파티션 컬럼"
        )
    schema = convert_oracle_to_bigquery_schema(
        owner=owner, table_name=oracle_table, partition_field=partition_field_schema
    )

    dataset_ref = bigquery.DatasetReference(bq_project, bq_dataset)
    table_ref = dataset_ref.table(bq_table_name)
    table = bigquery.Table(table_ref, schema=schema)
    table.time_partitioning = bigquery.TimePartitioning(
        type_=bigquery.TimePartitioningType.DAY,
        field=partition_field,
    )

    bq = get_bigquery_client(project=bq_project)

    result = bq.create_table(table, exists_ok=True)
    print(
        "Created table {}, partitioned on column {}".format(
            result.table_id, result.time_partitioning.field
        )
    )

    bq.close()


def get_bigquery_load_job_config(
    schema, write_disposition, partition_type=None, partition_field=None
):
    job_config = bigquery.LoadJobConfig()
    job_config.schema = schema
    job_config.write_disposition = write_disposition
    job_config.create_disposition = "CREATE_IF_NEEDED"
    job_config.source_format = bigquery.SourceFormat.PARQUET
    job_config.autodetect = False

    if partition_type and partition_field:
        job_config.time_partitioning = bigquery.TimePartitioning(
            type_=partition_type, field=partition_field
        )

    return job_config


def generate_oracle_select_query(owner, table_name, where=None):
    oracle_schema = get_oracle_schema(owner, table_name)[
        ["schema_name", "table_name", "column_name", "data_type_name"]
    ]

    if len(oracle_schema.index) == 0:
        return False

    col_list = []
    table = ""

    for idx, row in oracle_schema.iterrows():
        table = row["table_name"].strip()
        column = row["column_name"].strip()
        # schema = row['schema_name'].strip()
        # oc_type = row['data_type_name'].strip()

        enc_type = get_enc_type(owner, table, column)
        if enc_type == "AES":
            col_list.append(f"AES.DEC({column}, 'emrt') AS {column}")
        elif enc_type == "CUB":
            col_list.append(f"XX1.DEC_VARCHAR2_SEL({column}, 10, 'SSNF') AS {column}")
        else:
            col_list.append(column)

    query = "SELECT " + ", ".join(col_list) + " FROM " + owner + "." + table

    if where:
        query += f" WHERE {where}"

    return query


def oracle_to_bq_overwrite_table(
    owner,
    table,
    dataset,
    bq_table=None,
    bq_project=GCP_PROJECT,
    partition_column=None,
    partition_type=None,
    partition_field=None,
    query=None,
):
    bq_table = bq_table if bq_table is not None else table.lower()
    query = query if query is not None else generate_oracle_select_query(owner, table)
    pdf = oracle_to_pandas(owner, table, sql=query)

    schema = ""
    if partition_column and partition_field:
        pdf[partition_field] = pd.to_datetime(
            pdf[partition_column.lower()], format="%Y-%m-%d %H:%M:%S"
        ).dt.date
        partition_schema_field = bigquery.SchemaField(
            partition_field, "DATE", mode="REQUIRED", description="파티션 컬럼"
        )
        schema = convert_oracle_to_bigquery_schema(owner, table, partition_schema_field)
    else:
        schema = convert_oracle_to_bigquery_schema(owner, table)

    if schema:
        bq = get_bigquery_client(project=bq_project)
        t_ref = bigquery.TableReference(
            bigquery.DatasetReference(bq_project, dataset), bq_table
        )
        jc = get_bigquery_load_job_config(
            schema, "WRITE_TRUNCATE", partition_type, partition_field
        )
        j = bq.load_table_from_dataframe(pdf, t_ref, job_config=jc)
        j.result()

        bq.close()
